# Copyright 2014 Scratchwork Development LLC. All rights reserved.

class Prefab
  attr :notifications, :notification_callbacks

  def initialize
    @notification_callbacks = {}
    @thick_border_rts = {}
    @flashes ||= {}
    @notifications ||= {}
    @dismiss_queue ||= []
  end

  def queue_notification id, text, text_color: nil
    return if @notifications.any? { |k, v| v.text == text }
    $args.outputs.a11y[id] = { a11y_id: id,
                               a11y_text: text,
                               a11y_trait: :notification }
    @notifications[id] = {
      id: id,
      text: text,
      at: Kernel.tick_count + 15,
      text_color: text_color || white
    }
  end

  def notification_duration for_notification
    if for_notification.id == :story_line
      6.seconds
    elsif for_notification.id == :water
      2.seconds
    elsif for_notification.id == :food
      2.seconds
    elsif for_notification.id == :no_supplies
      2.seconds
    else
      3.seconds
    end
  end

  def notification_spline
    @notification_spline ||= [
      [0.0, 0.66, 1.0, 1.0],
      [1.0, 1.0,  1.0, 1.0],
      [1.0, 1.0,  1.0, 1.0],
      [1.0, 1.0,  1.0, 1.0],
      [1.0, 1.0,  0.33, 0.0]
    ]
  end

  def clear_notifications
    @notifications = {}
  end

  def dismiss_all_notifications
    @notifications.each do |k, v|
      @dismiss_queue << v
    end
    @notifications = {}
  end

  def dismiss_current_notification
    current_notification = @notifications.values.first
    return if !current_notification
    @dismiss_queue << current_notification
    @notifications.delete current_notification.id
  end

  def notification_h
    192
  end

  def tick_notifications
    key, first_notification = @notifications.first

    if first_notification && first_notification.at.elapsed_time > notification_duration(first_notification)
      @notifications.delete key
      @notification_callbacks.each do |k, v|
        v.call key, first_notification
      end
    end

    @notifications.each do |k, v|
      v.at = Kernel.tick_count if v != first_notification
    end

    @dismiss_queue.each do |v|
      v.at ||= Kernel.tick_count
      v.offset_y ||= 0
      v.offset_y = v.offset_y.lerp(Grid.allscreen_h - Grid.h + 200, 0.1).round(2)
    end

    @dismiss_queue.reject! do |v|
      v.at.elapsed_time > notification_duration(v) || notification_rect(v).y >= Grid.allscreen_h
    end
  end

  def tick_flashes
    to_delete = []
    @flashes.each do |k, v|
      v[:a] ||= 255
      v[:a] -= v[:rate] * 100
      if v[:a] < 0
        to_delete << k
      end
    end

    to_delete.each { |k| @flashes.delete(k) }
  end

  def color_with_perc_towards r, g, b, perc, t
    {
      r: (r + (perc * (t.r - r))).round,
      g: (g + (perc * (t.g - g))).round,
      b: (b + (perc * (t.b - b))).round
    }
  end

  def notification_rect notification
    perc = Easing.ease_spline notification.at,
                              Kernel.tick_count,
                              notification_duration(notification),
                              notification_spline
    shift_up_y = notification.offset_y || 0
    h = if Grid.landscape?
          notification_h
        else
          notification_h + Grid.allscreen_offset_y
        end
    {
      x: Grid.x,
      y: Grid.h - notification_h + 4 + shift_up_y,
      w: Grid.w,
      h: h,
    }
  end

  def notification_prefab notification
    perc = Easing.ease_spline notification.at, Kernel.tick_count, notification_duration(notification), notification_spline
    shift_up_y = notification.offset_y || 0
    a = perc * 255
    return if !notification.text
    lines = notification.text.wrapped_lines(38)
    background_solid = notification_rect(notification).merge!(path: :solid, **black, a: a)
    labels = lines.map_with_index do |l, i|
      {
        x: Grid.w / 2,
        y: Grid.h - 96 + shift_up_y,
        text: l.strip,
        font: "fonts/text.ttf",
        size_px: 34,
        anchor_x: 0.5,
        anchor_y: lines.length.idiv(2) * -0.50 + i,
        **notification.text_color,
        a: a
      }
    end
    [
      background_solid,
      labels
    ]
  end

  def notification_primitives
    current_notification = @notifications.values.first
    all_primitives = []

    if current_notification && current_notification.at <= Kernel.tick_count
      all_primitives << notification_prefab(current_notification)
    end

    @dismiss_queue.each do |d|
      all_primitives << notification_prefab(d)
    end

    all_primitives
  end

  def render_flash_primitives args
    @flashes.map do |k, f|
      if f.rt_name
        args.outputs[f.rt_name].primitives << {
          x: f[:x],
          y: f[:y],
          w: f[:w],
          h: f[:h],
          r: f[:r],
          g: f[:g],
          b: f[:b],
          a: f[:a],
          anchor_x: f[:anchor_x],
          anchor_y: f[:anchor_y],
          path: :solid
        }
      else
        args.outputs.primitives << {
          x: f[:x],
          y: f[:y],
          w: f[:w],
          h: f[:h],
          r: f[:r],
          g: f[:g],
          b: f[:b],
          a: f[:a],
          anchor_x: f[:anchor_x],
          anchor_y: f[:anchor_y],
          path: :solid
        }
      end
    end.to_a
  end

  def queue_flash id, x, y, w, h, r, g, b, rate, rt_name: nil, a: 255, anchor_x: 0, anchor_y: 0
    @flashes[id] = {
      x: x.round(0),
      y: y.round(0),
      w: w.round(0),
      h: h.round(0),
      r: r,
      g: g,
      b: b,
      a: a,
      rate: rate,
      path: :solid,
      anchor_x: anchor_x,
      anchor_y: anchor_y,
      rt_name: rt_name
    }
  end

  def thick_border args, b
    return b if b && b.path == :pixel
    x = b.x
    y = b.y
    w = b.w
    h = b.h
    @thick_border_rts ||= {}
    @thick_border_rts[w] ||= {}
    @thick_border_rts[w][h] ||= {}
    rt_name = "thick_border_#{w}_#{h}"
    @thick_border_rts[w][h] = rt_name
    args.outputs[rt_name].transient!
    args.outputs[rt_name].w = w
    args.outputs[rt_name].h = h
    args.outputs[rt_name].solids << {
      x: 0, y: 0, w: 2, h: h, r: 255, g: 255, b: 255
    }

    args.outputs[rt_name].solids << {
      x: w - 2, y: 0, w: 2, h: h, r: 255, g: 255, b: 255
    }

    args.outputs[rt_name].solids << {
      x: 2, y: 0, w: w - 2, h: 2, r: 255, g: 255, b: 255
    }

    args.outputs[rt_name].solids << {
      x: 2, y: h - 2, w: w - 2, h: 2, r: 255, g: 255, b: 255
    }

    color = if b.r
              { r: b.r, g: b.g, b: b.b }
            else
              black
            end

    { x: x, y: y, w: w, h: h, path: @thick_border_rts[w][h], **color, metadata: b.metadata }
  end

  def bright_white
    # if GTK.platform?(:mobile)
      { r: 255, g: 255, b: 255 }
    # else
    #   { r: 102 + 19 * 3, g: 108 + 19 * 3, b: 60 + 19 * 3 }
    # end
  end

  def white
    # if GTK.platform?(:mobile)
      { r: 255, g: 255, b: 255 }
    # else
    #   { r: 102, g: 108, b: 60 }
    # end
  end

  def really_light_gray_2
    # if GTK.platform?(:mobile)
      { r: 210, g: 210, b: 210 }
    # else
    #   { r: 102 - 19 * 1, g: 108 - 17 * 1, b: 60 - 7 * 1 }
    # end
  end

  def really_light_gray
    # if GTK.platform?(:mobile)
      { r: 200, g: 200, b: 200 }
    # else
    #   { r: 102 - 19 * 1, g: 108 - 17 * 1, b: 60 - 7 * 1 }
    # end
  end

  def kindof_light_gray
    { r: 128, g: 128, b: 128 }
  end

  def light_gray
    # if GTK.platform?(:mobile)
      { r: 80, g: 80, b: 80 }
    # else
    #   { r: 102 - 19 * 2, g: 108 - 17 * 2, b: 60 - 7 * 2 }
    # end
  end

  def gray
    # if GTK.platform?(:mobile)
      { r: 80, g: 80, b: 80 }
    # else
    #   { r: 102 - 19 * 3, g: 108 - 17 * 3, b: 60 - 7 * 3 }
    # end
  end

  def dark_gray
    # if GTK.platform?(:mobile)
      { r: 30, g: 30, b: 30 }
    # else
    #   { r: 102 - 19 * 4, g: 108 - 17 * 4, b: 60 - 7 * 4 }
    # end
  end

  def black
    # if GTK.platform?(:mobile)
      { r: 0, g: 0, b: 0 }
    # else
    #   { r: 6, g: 25, b: 25 }
    # end
  end

  def dark_black
    # if GTK.platform?(:mobile)
      { r: 0, g: 0, b: 0 }
    # else
    #   { r: 6, g: 15, b: 15 }
    # end
  end

  def button args, h
    return nil if !h
    color = if h.r
              { r: h.r, g: h.g, b: h.b, a: h.a }
            else
              black
            end

    bg_prefab = if h.bg_color
                  h.slice(:x, :y, :w, :h).merge(**h.bg_color, path: :solid)
                else
                  nil
                end

    size_px = h.size_px || 30

    center = if h.center
               h.center
             else
               Geometry.rect_props(h).center
             end

    tb = thick_border(args, h).merge(color)

    a11y_data = {}
    if GTK.a11y_enabled? && h.a11y_id
      tb = { **tb, a11y_trait: :button, **h.slice(:a11y_id, :a11y_text) }
    end

    [
      bg_prefab,
      center.merge(text: h.text, font: "fonts/text.ttf", size_px: size_px, anchor_x: 0.5, anchor_y: 0.5, **color),
      tb
    ]
  end

  def progress_bar args, pb
    return nil if !pb
    color = pb.text_color || dark_black
    size_px = pb.text_size_px || 28

    labels = if pb.text && pb.text_2
               [
                 { x: pb.x + pb.w / 2,
                   y: pb.y + pb.h / 2,
                   anchor_x: 0.5,
                   anchor_y: 0.15,
                   size_px: size_px,
                   primitive_marker: :label,
                   font: "fonts/text.ttf",
                   text: pb.text,
                   **color },
                 { x: pb.x + pb.w / 2,
                   y: pb.y + pb.h / 2,
                   anchor_x: 0.5,
                   anchor_y: 1.0,
                   size_px: size_px,
                   primitive_marker: :label,
                   font: "fonts/text.ttf",
                   text: pb.text_2,
                   **color }
               ]
             else
               [
                 { x: pb.x + pb.w / 2,
                   y: pb.y + pb.h / 2,
                   anchor_x: 0.5,
                   anchor_y: 0.5,
                   size_px: size_px,
                   primitive_marker: :label,
                   font: "fonts/text.ttf",
                   text: pb.text,
                   **color }
               ]
             end

    perc_bg = pb.merge(w: pb.w * pb.perc, path: :pixel, primitive_marker: :sprite, **really_light_gray)
    if perc_bg[:a11y_id]
      perc_bg.delete(:a11y_id)
      perc_bg.delete(:a11y_trait)
      perc_bg.delete(:a11y_text)
    end
    perc_border = thick_border(args, pb.merge({ w: (pb.w * pb.perc).to_i, **dark_black }))

    if perc_border[:a11y_id]
      perc_bg.delete(:a11y_id)
      perc_bg.delete(:a11y_trait)
      perc_bg.delete(:a11y_text)
    end

    border_prim = if GTK.a11y_enabled?
                    a11y = pb.slice(:a11y_id, :a11y_trait, :a11y_text)
                    thick_border(args, pb.merge(dark_black)).merge(a11y)
                  else
                    thick_border(args, pb.merge(dark_black))
                  end

    [
      perc_bg,
      perc_border,
      *labels,
      border_prim
    ]
  end

  def labels lines:, rect:, wrap_length: 34, size_px: 30, font: "fonts/text.ttf", a11y_id: nil
    message_rect = Layout.rect(**rect)
    text_with_location = []

    starting_x = message_rect.x
    starting_y = message_rect.y

    anchor_increment = 0
    lines.each_with_index do |t, j|
      multi_line = t.strip.wrapped_lines(wrap_length).reject { |l| l.strip.length == 0 }
      multi_line.each_with_index do |line, i|
        anchor_increment += 1.2
        text_with_location << { x: message_rect.x,
                                y: message_rect.center.y,
                                anchor_increment: anchor_increment,
                                font: font,
                                size_px: size_px,
                                text: line.strip }
      end
      anchor_increment += 1.0
    end

    l = text_with_location.length

    text_with_location.each_with_index do |t, i|
      t.anchor_y = -(l / 2) + t.anchor_increment - 0.5
    end

    if a11y_id && text_with_location[0] && GTK.a11y_enabled?
    text_with_location[0].a11y_id = a11y_id
    text_with_location[0].a11y_trait = :label
    text_with_location[0].a11y_text = (text_with_location.map { |s| s.text }.join " \n ")
    end

    text_with_location
  end
end
