# Copyright 2014 Scratchwork Development LLC. All rights reserved.

class Scene
  attr_gtk

  attr :game, :prefab, :metadata, :selection

  def initialize
    @selection_activated_at = nil
  end

  def game= value
    @game = value
  end

  def id
    raise "#{self.class} must implement #id"
  end

  def tick
    outputs.primitives << { x: Grid.w / 2,
                            y: Grid.h / 2,
                            text: "Implement tick in #{self.class}",
                            anchor_x: 0.5,
                            anchor_y: 0.5 }
  end

  def will_appear_base
    will_appear
    if GTK.a11y_enabled?
      args.outputs.a11y_clear_pending_notifications!
      target = will_appear_a11y_target
      if target
        args.outputs.a11y[:notify_will_appear] = {
          a11y_trait: :notification,
          a11y_notification_target: target
        }
      end
    end
  end

  def will_appear
    raise "#{self.class} must implement #will_appear"
  end

  def will_appear_a11y_target
    nil
  end

  def long_lines_split line, length
    line.wrapped_lines length
  end

  def make_new_lines word, rest, length, aggregate
    if word.nil?
      return aggregate
    elsif rest[0].nil?
      aggregate << word
      return aggregate
    elsif (word + " " + rest[0]).length > length
      aggregate << word
      return make_new_lines rest[0], rest[1..-1], length, aggregate
    elsif (word + " " + rest[0]).length <= length
      next_word = (word + " " + rest[0])
      return make_new_lines next_word, rest[1..-1], length, aggregate
    else
      raise "#{word} is too damn long to fit in length of #{length}."
    end
  end

  def text_color
    prefab.dark_black
  end

  def stop_music
    if audio["bg_music"]
      bg = audio["bg_music"]
      audio["bg_music_fade"] = { input: bg.input, looping: true, gain: bg.gain, pitch: bg.pitch, paused: false, playtime: bg.playtime }
    end

    audio["bg_music"] = nil
  end

  def queue_sound path
    @one_time_id ||= 1
    audio[@one_time_id] = { input: path, gain: 1.0, pitch: 1.0, paused: false, looping: false }
    @one_time_id += 1
  end

  def queue_music path, repeat, initial_gain = 0.0
    return if audio["bg_music"] && audio["bg_music"][:input] == path

    if audio["bg_music"]
      bg = audio["bg_music"]
      bg_gain = bg.gain - initial_gain
      bg_gain = bg_gain.clamp(0.0, 1.0)
      audio["bg_music_fade"] = { input: bg.input,
                                 looping: true,
                                 gain: bg_gain.to_f,
                                 pitch: bg.pitch,
                                 paused: false,
                                 playtime: bg.playtime }
    end

    audio["bg_music"] = { input: path,
                          looping: true,
                          gain: initial_gain,
                          pitch: 1.0,
                          paused: false,
                          looping: repeat }
  end

  def buttons
    r = button_ids.map do |b|
      { id: b, rect: self.send(b) }
    end
    r.reject! { |h| !h.rect }
    r
  end

  def progress_bars
    r = progress_bar_ids.map do |b|
      { id: b, rect: self.send(b) }
    end
    r.reject! { |h| !h.rect }
    r
  end

  def button_ids
    raise "#{self.class} must implement button_ids"
  end

  def render_buttons
    outputs.primitives << buttons.map { |b| prefab.button(args, b.rect) }
  end

  def render_progress_bars
    outputs.primitives << progress_bars.map do |b|
      prefab.progress_bar(args, b.rect)
    end
  end

  def calc_buttons
    return if buttons.length == 0

    if inputs.last_active == :mouse && inputs.mouse.click
      clicked_button = buttons.find { |b| inputs.mouse.inside_rect? b.rect }
      if clicked_button
        queue_flash :button_clicked,
                    clicked_button.rect.x,
                    clicked_button.rect.y,
                    clicked_button.rect.w,
                    clicked_button.rect.h,
                    prefab.black.r,
                    prefab.black.g,
                    prefab.black.b,
                    1.0/10.0


        if clicked_button.metadata && clicked_button.metadata.callback
          clicked_button.metadata.callback.call
        else
          m = "#{clicked_button.id}_selected".to_sym
          if !self.respond_to? m
            raise <<-S
#{self.class} must implement:

  def #{m}
  end
S
          end
          self.send m
        end
      end
    elsif inputs.last_active == :controller || inputs.last_active == :keyboard
    end
  end

  def available_selections_visible
    return [] if inputs.last_active != :controller && inputs.last_active != :keyboard && !GTK.a11y_enabled?
    available_selections.find_all { |b| (b.rect.a || 255) != 0 }
  end

  def available_selections
    return [] if inputs.last_active != :controller && inputs.last_active != :keyboard
    buttons + progress_bars
  end

  def progress_bar_ids
    []
  end

  def disable_selection_prefab!
    @selection_prefab_disabled = true
  end

  def enable_selection_prefab!
    @selection_prefab_disabled = false
  end

  def set_selection! id
    if !id
      @selection = nil
    else
      @selection = { current: available_selections.find { |b| b.id == id } }
      if @selection.current && GTK.a11y_enabled?

      end
    end
  end

  def input_cancel?
    return false if inputs.last_active != :controller && inputs.last_active != :keyboard
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.escape
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.b
    end
  end

  def calc_selection_prefab
    if inputs.last_active != :controller && inputs.last_active != :keyboard
      @selection_activated_at = nil
      return
    else
      @selection_activated_at ||= Kernel.tick_count
    end

    return if @selection_prefab_disabled

    if @selection_activated_at == Kernel.tick_count
      @selection = {
        current: (available_selections.sort_by { |s| -s.rect.y }).first,
      }
      return
    end

    if input_cancel? && prefab.notifications.length > 0
      prefab.dismiss_current_notification
      return
    end

    if !@selection
      @selection = {
        current: (available_selections.sort_by { |s| -s.rect.y }).first,
      }
      return
    end

    if @selection
      current_button = available_selections.find { |b| b.id == @selection.current.id }
      if !current_button
        @selection.current = available_selections.first
      end
    end

    return if !@selection.current

    up_down = if input_down?
                -1
              elsif input_up?
                1
              else
                0
              end

    left_right = if input_left?
                   -1
                 elsif input_right?
                   1
                 else
                   0
                 end


    if up_down != 0 || left_right != 0
      next_rect = Geometry.rect_navigate(
        rect: @selection.current,
        rects: available_selections_visible,
        up_down: up_down,
        left_right: left_right,
        wrap_x: true,
        wrap_y: true,
        using: lambda { |r| r.rect }
      )

      if next_rect && next_rect != @selection.current.rect
        @selection.current = next_rect
        if respond_to? :selection_changed
          selection_changed @selection.current
        end
      end
    end

    if input_accept?
      if @selection.current.metadata && @selection.current.metadata.callback
        @selection.current.metadata.callback.call

        queue_flash :button_clicked,
                    @selection.current.rect.x,
                    @selection.current.rect.y,
                    @selection.current.rect.w,
                    @selection.current.rect.h,
                    prefab.black.r,
                    prefab.black.g,
                    prefab.black.b,
                    1.0/10.0
      else
        m = "#{@selection[:current][:id]}_selected"
        if !self.respond_to? m
          raise <<-S
#{self.class} must implement:

  def #{m}
  end
S
        end
        self.send m
      end
    end

    if inputs.controller_one.key_down.b
      last_selection = back_button_id || available_selections_visible.sort_by { |s| s.rect.y }.first
      if last_selection && last_selection.id
        set_selection! last_selection.id
      end
    end
  end

  def back_button_id
  end

  def input_accept?
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.enter || inputs.keyboard.key_down.space
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.a
    else
      false
    end
  end

  def input_down?
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.down || inputs.keyboard.key_down.s_scancode
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.down
    else
      false
    end
  end

  def input_left?
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.left || inputs.keyboard.key_down.a_scancode
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.left
    else
      false
    end
  end

  def input_right?
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.right || inputs.keyboard.key_down.d_scancode
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.right
    else
      false
    end
  end

  def input_up?
    if inputs.last_active == :keyboard
      inputs.keyboard.key_down.up || inputs.keyboard.key_down.w_scancode
    elsif inputs.last_active == :controller
      inputs.controller_one.key_down.up
    else
      false
    end
  end

  def selection_prefab
    return nil if state.queued_scene
    return if inputs.last_active != :controller && inputs.last_active != :keyboard
    return if !@selection
    return if !@selection.current
    return if @selection_prefab_disabled

    selection_rect = @selection.current.rect.slice(:x, :y, :w, :h)
    selection_rect.x -= 3
    selection_rect.y -= 3
    selection_rect.w += 6
    selection_rect.h += 6

    selection_rect_2 = @selection.current.rect.slice(:x, :y, :w, :h)
    selection_rect_2.x -= 5
    selection_rect_2.y -= 5
    selection_rect_2.w += 10
    selection_rect_2.h += 10

    # selection_rect_3 = @selection.current.rect.slice(:x, :y, :w, :h)
    # selection_rect_3.x -= 6
    # selection_rect_3.y -= 6
    # selection_rect_3.w += 12
    # selection_rect_3.h += 12

    [
      prefab.thick_border(args, selection_rect),
      prefab.thick_border(args, selection_rect_2),
      selection_rect_2.merge(r: 0, g: 255, b: 0, a: 32, path: :solid)
      # prefab.thick_border(args, selection_rect_3)
    ]
  end

  def calc_progress_bars
    return if !inputs.mouse.click
    clicked_progress_bar = progress_bars.find { |b| inputs.mouse.inside_rect? b.rect }
    if clicked_progress_bar
      queue_flash :button_clicked,
                  clicked_progress_bar.rect.x,
                  clicked_progress_bar.rect.y,
                  clicked_progress_bar.rect.w,
                  clicked_progress_bar.rect.h,
                  prefab.black.r,
                  prefab.black.g,
                  prefab.black.b,
                  1.0/10.0

      m = "#{clicked_progress_bar.id}_selected".to_sym
      if !self.respond_to? m
        raise <<-S
#{self.class} must implement:

  def #{m}
  end
S
      end
      self.send m
    end
  end

  def create_progress_bar args
    { x: args[:x].to_i,
      y: args[:y].to_i,
      w: args[:w].to_i,
      h: args[:h].to_i,
      text_color: args[:text_color],
      text: args[:text],
      text_2: args[:text_2],
      perc: args[:perc],
      rate: args[:rate].to_i,
      a11y_id: args[:a11y_id],
      a11y_text: args[:a11y_text],
      a11y_trait: args[:a11y_trait] }
  end

  def queue_flash id, x, y, w, h, r, g, b, rate, anchor_x: 0, anchor_y: 0
    @prefab.queue_flash id, x, y, w, h, r, g, b, rate, anchor_x: 0, anchor_y: 0
  end

  def queue_notification id, text, text_color: nil
    @prefab.queue_notification id, text, text_color: text_color
  end

  def clear_notifications
    @prefab.clear_notifications
  end

  def changing_scene?
    state.next_scene || state.queued_scene
  end

  def controller_instructions accept_char
    default_controller_instructions
  end

  def keyboard_instructions
    default_keyboard_instructions
  end

  def default_keyboard_instructions
    "[ARROW KEYS/WASD]: move cursor. [SPACE/ENTER]: select."
  end

  def default_controller_instructions
    "[DIRECTIONAL]: move cursor. [#{controller_instructions_accept_char}]: select."
  end

  def controller_instructions_accept_char
    if inputs.controller_one.name.downcase.start_with? "ps"
      "X"
    else
      "A"
    end
  end

  def instructions_prefab
    return nil if Grid.portrait?

    if inputs.last_active == :controller && inputs.controller_one
      accept_char = controller_instructions_accept_char

      text = controller_instructions accept_char
      return nil if !text

      return Layout.rect(row: 12, col: 0, w: 24, h: 0)
                   .center
                   .merge(text: text, font: "fonts/text.ttf", anchor_y: 1.5, anchor_x: 0.5)
    elsif inputs.last_active == :keyboard
      text = keyboard_instructions
      return nil if !text

      return Layout.rect(row: 12, col: 0, w: 24, h: 0)
                   .center
                   .merge(text: text, font: "fonts/text.ttf", anchor_y: 1.5, anchor_x: 0.5)
    else
      return nil
    end
  end
end
